// Copyright 2021 The Chromium Authors
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

import './shimless_rma_shared.css.js';
import './base_page.js';

import {I18nBehavior, I18nBehaviorInterface} from 'chrome://resources/ash/common/i18n_behavior.js';
import {mixinBehaviors, PolymerElement} from 'chrome://resources/polymer/v3_0/polymer/polymer_bundled.min.js';

import {getShimlessRmaService} from './mojo_interface_provider.js';
import {getTemplate} from './onboarding_wp_disable_complete_page.html.js';
import {ShimlessRmaServiceInterface, StateResult, WriteProtectDisableCompleteAction} from './shimless_rma.mojom-webui.js';
import {enableNextButton, focusPageTitle} from './shimless_rma_util.js';

/** @type {!Object<WriteProtectDisableCompleteAction, string>} */
const disableActionTextKeys = {
  [WriteProtectDisableCompleteAction.kSkippedAssembleDevice]:
      'wpDisableReassembleNowText',
  [WriteProtectDisableCompleteAction.kCompleteAssembleDevice]:
      'wpDisableReassembleNowText',
  [WriteProtectDisableCompleteAction.kCompleteKeepDeviceOpen]:
      'wpDisableLeaveDisassembledText',
};

/**
 * @fileoverview
 * 'onboarding-wp-disable-complete-page' notifies the user that manual HWWP
 * disable was successful, and what steps must be taken next.
 */

/**
 * @constructor
 * @extends {PolymerElement}
 * @implements {I18nBehaviorInterface}
 */
const OnboardingWpDisableCompletePageBase =
    mixinBehaviors([I18nBehavior], PolymerElement);

/** @polymer */
export class OnboardingWpDisableCompletePage extends
    OnboardingWpDisableCompletePageBase {
  static get is() {
    return 'onboarding-wp-disable-complete-page';
  }

  static get template() {
    return getTemplate();
  }

  static get properties() {
    return {
      /** @protected */
      actionString: {
        type: String,
        computed: 'getActionString(action)',
      },
    };
  }

  /** @override */
  ready() {
    super.ready();
    enableNextButton(this);

    focusPageTitle(this);
  }

  constructor() {
    super();
    /** @private {ShimlessRmaServiceInterface} */
    this.shimlessRmaService = getShimlessRmaService();
    /** @private {WriteProtectDisableCompleteAction} */
    this.action = WriteProtectDisableCompleteAction.kUnknown;

    this.shimlessRmaService.getWriteProtectDisableCompleteAction().then(
        (res) => {
          if (res) {
            this.action = res.action;
          }
        });
  }

  /**
   * @return {string}
   * @protected
   */
  getActionString() {
    return (this.action === WriteProtectDisableCompleteAction.kUnknown ||
            this.action === WriteProtectDisableCompleteAction.kCompleteNoOp) ?
        '' :
        this.i18n(disableActionTextKeys[this.action]);
  }

  /** @return {!Promise<!{stateResult: !StateResult}>} */
  onNextButtonClick() {
    return this.shimlessRmaService.confirmManualWpDisableComplete();
  }

  /**
   * @return {string}
   * @protected
   */
  getVerificationIcon() {
    return (this.action === WriteProtectDisableCompleteAction.kUnknown ||
            this.action === WriteProtectDisableCompleteAction.kCompleteNoOp) ?
        '' :
        'shimless-icon:check';
  }
}

customElements.define(
    OnboardingWpDisableCompletePage.is, OnboardingWpDisableCompletePage);
